#include "stdafx.h"

#define Min(x,y) (x < y ? x : y) 

#define PARTFILE_VERSION				0xe0
#define PARTFILE_SPLITTEDVERSION		0xe1
#define PARTFILE_INTERNAL				0x01020000

#define FSTYPE_FAT	0
#define FSTYPE_NTFS	1

typedef enum _EPART_FILE_FORMAT {
	PMT_UNKNOWN = 0,
	PMT_DEFAULTOLD,
	PMT_SPLITTED,
	PMT_NEWOLD,
	PMT_SHAREAZA,
	PMT_BADFORMAT	
} EPART_FILE_FORMAT;

#define FT_FILENAME				 0x01	// <string>
#define TAG_NAME				"\x01"	// <string>
#define FT_FILESIZE				 0x02	// <uint32>
#define TAG_SIZE				"\x02"	// <uint32>
#define FT_FILETYPE				 0x03	// <string>
#define TAG_TYPE				"\x03"	// <string>
#define FT_FILEFORMAT			 0x04	// <string>
#define TAG_FORMAT				"\x04"	// <string>
#define FT_LASTSEENCOMPLETE		 0x05	// <uint32>
#define TAG_COLLECTION			"\x05"
#define	TAG_PART_PATH			"\x06"	// <string>
#define	TAG_PART_HASH			"\x07"
#define FT_TRANSFERED			 0x08	// <uint32>
#define	TAG_COPIED				"\x08"	// <uint32>
#define FT_GAPSTART				 0x09	// <uint32>
#define	TAG_GAP_START			"\x09"	// <uint32>
#define FT_GAPEND				 0x0A	// <uint32>
#define	TAG_GAP_END				"\x0A"	// <uint32>
#define	TAG_DESCRIPTION			"\x0B"	// <string>
#define	TAG_PING				"\x0C"
#define	TAG_FAIL				"\x0D"
#define	TAG_PREFERENCE			"\x0E"
#define TAG_PORT				"\x0F"
#define TAG_IP_ADDRESS			"\x10"
#define TAG_VERSION				"\x11"	// <string>
#define FT_PARTFILENAME			 0x12	// <string>
#define TAG_TEMPFILE			"\x12"	// <string>
//#define FT_PRIORITY			 0x13	// Not used anymore
#define TAG_PRIORITY			"\x13"	// <uint32>
#define FT_STATUS				 0x14	// <uint32>
#define TAG_STATUS				"\x14"	// <uint32>
#define FT_SOURCES				 0x15	// <uint32>
#define TAG_AVAILABILITY		"\x15"	// <uint32>
#define FT_PERMISSIONS			 0x16	// <uint32>
#define TAG_QTIME				"\x16"
//#define FT_ULPRIORITY			 0x17	// Not used anymore
#define TAG_PARTS				"\x17"
#define FT_DLPRIORITY			 0x18	// Was 13
#define FT_ULPRIORITY			 0x19	// Was 17
#define FT_KADLASTPUBLISHKEY	 0x20	// <uint32>
#define FT_KADLASTPUBLISHSRC	 0x21	// <uint32>
#define	FT_FLAGS				 0x22	// <uint32>
#define	FT_DL_ACTIVE_TIME		 0x23	// <uint32>
#define	FT_CORRUPTEDPARTS		 0x24	// <string>
#define FT_DL_PREVIEW            0x25
#define FT_AICH_HASH			 0x27
#define	FT_COMPLETE_SOURCES		 0x30	// nr. of sources which share a complete version of the associated file (supported by eserver 16.46+)
// statistic
#define FT_ATTRANSFERED			 0x50	// <uint32>
#define FT_ATREQUESTED			 0x51	// <uint32>
#define FT_ATACCEPTED			 0x52	// <uint32>
#define FT_CATEGORY				 0x53	// <uint32>
#define	FT_ATTRANSFEREDHI		 0x54	// <uint32>
#define	TAG_MEDIA_ARTIST		"\xD0"	// <string>
#define	 FT_MEDIA_ARTIST		 0xD0	// <string>
#define	TAG_MEDIA_ALBUM			"\xD1"	// <string>
#define	 FT_MEDIA_ALBUM			 0xD1	// <string>
#define	TAG_MEDIA_TITLE			"\xD2"	// <string>
#define	 FT_MEDIA_TITLE			 0xD2	// <string>
#define	TAG_MEDIA_LENGTH		"\xD3"	// <uint32> !!!
#define	 FT_MEDIA_LENGTH		 0xD3	// <uint32> !!!
#define	TAG_MEDIA_BITRATE		"\xD4"	// <uint32>
#define	 FT_MEDIA_BITRATE		 0xD4	// <uint32>
#define	TAG_MEDIA_CODEC			"\xD5"	// <string>
#define	 FT_MEDIA_CODEC			 0xD5	// <string>
#define TAG_BUDDYHASH			"\xF8"	// <string>
#define TAG_CLIENTLOWID			"\xF9"	// <uint32>
#define TAG_SERVERPORT			"\xFA"	// <uint16>
#define TAG_SERVERIP			"\xFB"	// <uint32>
#define TAG_SOURCEUPORT			"\xFC"	// <uint16>
#define TAG_SOURCEPORT			"\xFD"	// <uint16>
#define TAG_SOURCEIP			"\xFE"	// <uint32>
#define TAG_SOURCETYPE			"\xFF"	// <uint8>

#define	TAGTYPE_HASH			0x01
#define	TAGTYPE_STRING			0x02
#define	TAGTYPE_UINT32			0x03
#define	TAGTYPE_FLOAT32			0x04
#define	TAGTYPE_BOOL			0x05
#define	TAGTYPE_BOOLARRAY		0x06
#define	TAGTYPE_BLOB			0x07
#define	TAGTYPE_UINT16			0x08
#define	TAGTYPE_UINT8			0x09
#define	TAGTYPE_BSOB			0x0A
#define TAGTYPE_STR1			0x11
#define TAGTYPE_STR2			0x12
#define TAGTYPE_STR3			0x13
#define TAGTYPE_STR4			0x14
#define TAGTYPE_STR5			0x15
#define TAGTYPE_STR6			0x16
#define TAGTYPE_STR7			0x17
#define TAGTYPE_STR8			0x18
#define TAGTYPE_STR9			0x19
#define TAGTYPE_STR10			0x1A
#define TAGTYPE_STR11			0x1B
#define TAGTYPE_STR12			0x1C
#define TAGTYPE_STR13			0x1D
#define TAGTYPE_STR14			0x1E
#define TAGTYPE_STR15			0x1F
#define TAGTYPE_STR16			0x20

typedef struct _CTAG {
	BYTE Type;
	BYTE uName;
	TCHAR* sName;
	DWORD BlobSize;
	union {
		BYTE Valu8;
		WORD Valu16;
		DWORD Valu32;
		FLOAT Valfp32;
		TCHAR *ValStr;
		VOID *ValPtr;
	};
} CTAG;

typedef struct _CGAP {
	UINT64 Start, End;
} CGAP;

#define	SHORT_ED2K_STR			256
#define	SHORT_RAW_ED2K_MB_STR	(SHORT_ED2K_STR * 2)
#define	SHORT_RAW_ED2K_UTF8_STR	(SHORT_ED2K_STR * 4)

int GetX() {
	CONSOLE_SCREEN_BUFFER_INFO sb;
	HANDLE hOutput = GetStdHandle(STD_OUTPUT_HANDLE);
	GetConsoleScreenBufferInfo(hOutput, &sb);
	return sb.dwCursorPosition.X;
}

int GetY() {
	CONSOLE_SCREEN_BUFFER_INFO sb;
	HANDLE hOutput = GetStdHandle(STD_OUTPUT_HANDLE);
	GetConsoleScreenBufferInfo(hOutput, &sb);
	return sb.dwCursorPosition.Y;
}

void SetPosition(int x, int y) {
	COORD c;
	HANDLE hOutput = GetStdHandle(STD_OUTPUT_HANDLE);
	c.X = x;
	c.Y = y;
	SetConsoleCursorPosition(hOutput, c);
}

TCHAR* ReadString(BYTE* MetFileBuffer, DWORD *MetPtr, DWORD uLen) {
	TCHAR *Str = NULL;
	DWORD Idx;
	Str = (TCHAR*) LocalAlloc(LPTR, uLen + sizeof(CHAR));
	for (Idx = 0; Idx < uLen; Idx++) {
		Str[Idx] = *(MetFileBuffer + (*MetPtr) + Idx);
	}
	Str[Idx] = '\0';
	(*MetPtr) += Idx;
	return Str;
}

CTAG *CreateTag(BYTE* MetFileBuffer, DWORD *MetPtr) {
	DWORD Idx;
	CTAG *Tag = (CTAG*) LocalAlloc(LPTR, sizeof(CTAG));
	if (Tag == NULL) return NULL;

	Tag->Type = *(MetFileBuffer + (*MetPtr));
	(*MetPtr) ++;

	if (Tag->Type & 0x80) {
		Tag->Type &= 0x7F;
		Tag->uName = *(MetFileBuffer + (*MetPtr));
		(*MetPtr) ++;
	} else {
		UINT Length = *(WORD*) (MetFileBuffer + (*MetPtr));
		(*MetPtr) += sizeof(WORD);
		if (Length == 1) {
			Tag->uName = *(MetFileBuffer + (*MetPtr));
			(*MetPtr) ++;
		} else {
			Tag->uName = 0;
			Tag->sName = (TCHAR*) LocalAlloc(LPTR, Length + 1);
			for (Idx = 0; Idx < Length; Idx++) {
				Tag->sName[Idx] = *(MetFileBuffer + (*MetPtr) + Idx);
			}
			Tag->sName[Length] = '\0';
			(*MetPtr) += Length;
		}
	}
	Tag->BlobSize = 0;

	if (Tag->Type == TAGTYPE_STRING) {
		UINT uLen = *(WORD*) (MetFileBuffer + (*MetPtr));
		(*MetPtr) += sizeof(WORD);
		Tag->ValStr = ReadString(MetFileBuffer, MetPtr, uLen);
	} else if (Tag->Type == TAGTYPE_UINT32) {
		Tag->Valu32 = *(DWORD*) (MetFileBuffer + (*MetPtr));
		(*MetPtr) += sizeof(DWORD);
	} else if (Tag->Type == TAGTYPE_UINT16) {
		Tag->Valu32 = *(WORD*) (MetFileBuffer + (*MetPtr));
		(*MetPtr) += sizeof(WORD);
		Tag->Type = TAGTYPE_UINT32;
	} else if (Tag->Type == TAGTYPE_UINT8) {
		Tag->Valu8 = *(BYTE*) (MetFileBuffer + (*MetPtr));
		(*MetPtr) += sizeof(BYTE);
		Tag->Type = TAGTYPE_UINT32;
	} else if (Tag->Type == TAGTYPE_FLOAT32) {
		Tag->Valfp32 = *(FLOAT*) (MetFileBuffer + (*MetPtr));
		(*MetPtr) += sizeof(FLOAT);
	} else if (Tag->Type >= TAGTYPE_STR1 && Tag->Type <= TAGTYPE_STR16) {
		UINT Length = Tag->Type - TAGTYPE_STR1 + 1;
		Tag->ValStr = ReadString(MetFileBuffer, MetPtr, Length);
		Tag->Type = TAGTYPE_STRING;
	} else if (Tag->Type == TAGTYPE_HASH) {
		BYTE *Hash = (BYTE*) LocalAlloc(LPTR, 16 * sizeof(BYTE));
		for (Idx = 0; Idx < 16; Idx++) {
			Hash[Idx] = *(MetFileBuffer + (*MetPtr) + Idx);
		}
		(*MetPtr) += Idx;
		Tag->ValPtr = Hash;
	} else if (Tag->Type == TAGTYPE_BOOL) {
		(*MetPtr) ++;
	} else if (Tag->Type == TAGTYPE_BOOLARRAY) {
		WORD uLen = *(WORD*) (MetFileBuffer + (*MetPtr));
		(*MetPtr) += sizeof(WORD);
		(*MetPtr) += (uLen / 8) + 1;
	} else if (Tag->Type == TAGTYPE_BLOB) {
		Tag->BlobSize = *(DWORD*) (MetFileBuffer + (*MetPtr));
		(*MetPtr) += sizeof(DWORD);
		Tag->ValPtr = LocalAlloc(LPTR, Tag->BlobSize);
		for (Idx = 0; Idx < Tag->BlobSize; Idx++) {
			((BYTE*) Tag->ValPtr)[Idx] = *(MetFileBuffer + (*MetPtr));
		}
		(*MetPtr) += Idx;
	} else {
		printf("%s: Unkown Tag Type %x\n", __FUNCTION__, Tag->Type);
	}
	return Tag;
}

void DestroyTag(CTAG *Tag) {
	if (Tag->Type == TAGTYPE_BLOB) {
		LocalFree(Tag->ValPtr);
	} else if (Tag->Type == TAGTYPE_STRING) {
		LocalFree(Tag->ValStr);
	} else if (Tag->Type == TAGTYPE_HASH) {
		LocalFree(Tag->ValPtr);
	}
	if (Tag->sName) LocalFree(Tag->sName);
	LocalFree(Tag);
}

bool Rebuild(BYTE* MetFileBuffer, DWORD *MetPtr, char *Output) {
	HANDLE hFile;
	BYTE MetVersion;
	BOOL MetNewStyle;
	EPART_FILE_FORMAT MetType;
	DWORD MetDate;
	DWORD NumWrite;
	DWORD Idx;
	DWORD MetTagCount, TagIdx, MetOldPtr, MetGapTag = 0, MetFileSize = 0;

	hFile = CreateFile(Output, GENERIC_WRITE, 0, NULL, CREATE_NEW, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE) {
		return FALSE;
	}

	MetVersion = *(MetFileBuffer + (*MetPtr)++);
	WriteFile(hFile, &MetVersion, 1, &NumWrite, NULL);
	
	if (MetVersion != PARTFILE_VERSION && MetVersion != PARTFILE_SPLITTEDVERSION) {
		printf("Error: .met file version %#x invalid!\n", MetVersion);
		return 0;
	}

	MetNewStyle = (MetVersion == PARTFILE_SPLITTEDVERSION);
	MetType = MetNewStyle ? PMT_SPLITTED : PMT_DEFAULTOLD;

	if (!MetNewStyle) {
		DWORD Internal;
		*MetPtr = 24;
		Internal = *(DWORD*) (MetFileBuffer + (*MetPtr));
		
		SetFilePointer(hFile, 24, NULL, FILE_BEGIN);
		WriteFile(hFile, &Internal, sizeof(DWORD), &NumWrite, NULL);
		SetFilePointer(hFile, 1, NULL, FILE_BEGIN);

		*MetPtr = 1;
		if (Internal == PARTFILE_INTERNAL) {
			MetNewStyle = TRUE;
			MetType = PMT_NEWOLD;
		}
	}

	if (MetNewStyle) {
		DWORD Temp;
		Temp = *(DWORD*) (MetFileBuffer + *MetPtr);
		(*MetPtr) += sizeof(DWORD);

		WriteFile(hFile, &Temp, sizeof(DWORD), &NumWrite, NULL);

		if (Temp == 0) {
			goto LoadHashsetFromFile;
		} else {
			BYTE Hash[16];
			*MetPtr = 2;
			SetFilePointer(hFile, 1, NULL, FILE_BEGIN);

			MetDate = *(DWORD*) (MetFileBuffer + *MetPtr);
			(*MetPtr) += sizeof(DWORD);
			
			WriteFile(hFile, &MetDate, sizeof(DWORD), &NumWrite, NULL);

			printf("Hash Id: ");
			for (Idx = 0; Idx < 16; Idx++) {
				Hash[Idx] = *(MetFileBuffer + (*MetPtr) + Idx);
				printf("%.2x", Hash[Idx]);
			}
			WriteFile(hFile, Hash, 16, &NumWrite, NULL);
			MetPtr += Idx;
			printf("\n");
		}
	} else {
		BYTE Hash[16];
		WORD MetParts;
		DWORD PartIdx;

		MetDate = *(DWORD*) (MetFileBuffer + *MetPtr);
		(*MetPtr) += sizeof(DWORD);
		WriteFile(hFile, &MetDate, sizeof(DWORD), &NumWrite, NULL);

LoadHashsetFromFile:
		printf("Hash Id: ");
		for (Idx = 0; Idx < 16; Idx++) {
			Hash[Idx] = *(MetFileBuffer + (*MetPtr) + Idx);
			printf("%.2x", Hash[Idx]);
		}
		
		WriteFile(hFile, Hash, 16, &NumWrite, NULL);

		printf("\n");
		(*MetPtr) += Idx * sizeof(BYTE);
		MetParts = *(WORD*) (MetFileBuffer + *MetPtr);
		(*MetPtr) += sizeof(WORD);
		
		WriteFile(hFile, &MetParts, sizeof(WORD), &NumWrite, NULL);

		printf("Hashed Part Count: %d\n", MetParts);
		for (PartIdx = 0; PartIdx < MetParts; PartIdx++) {
//			printf("Hash Id for Part %d: ", PartIdx);
			for (Idx = 0; Idx < 16; Idx++) {
				Hash[Idx] = *(MetFileBuffer + (*MetPtr) + Idx);
//				printf("%.2x", Hash[Idx]);
			}
//			printf("\n");
			*(MetPtr) += Idx * sizeof(BYTE);
			WriteFile(hFile, Hash, 16, &NumWrite, NULL);
		}
	}
	MetTagCount = *(DWORD*) (MetFileBuffer + *MetPtr);
	(*MetPtr) += sizeof(DWORD);
	
	MetOldPtr = *MetPtr;
	for (TagIdx = 0; TagIdx < MetTagCount; TagIdx++) {
		CTAG *Tag = CreateTag(MetFileBuffer, MetPtr);
		if (Tag->uName == 0) {
			if (Tag->sName[0] == FT_GAPSTART) {
				MetGapTag++;
			}
			if(Tag->sName[0] == FT_GAPEND) {
				MetGapTag++;
			}
		} else if (Tag->uName == FT_FILESIZE) {
			MetFileSize = Tag->Valu32;
		}
		DestroyTag(Tag);
	}
	
	MetTagCount -= MetGapTag;
	WriteFile(hFile, &MetTagCount, sizeof(DWORD), &NumWrite, NULL);
	
	printf("Tag Count : %d\n", MetTagCount);
	
	*MetPtr = MetOldPtr;

	for (TagIdx = 0; TagIdx < MetTagCount + MetGapTag; TagIdx++) {
		CTAG *Tag = CreateTag(MetFileBuffer, MetPtr);
		if (Tag->uName == 0) {
			if (Tag->sName[0] == FT_GAPSTART) {
				/* ignore gaps */
				DestroyTag(Tag);
				continue;
			} else if (Tag->sName[0] == FT_GAPEND) {
				/* ignore gaps */
				DestroyTag(Tag);
				continue;
			}
		} else if (Tag->uName == FT_TRANSFERED) {
			Tag->Valu32 = MetFileSize;
		}

		if (Tag->sName == NULL) {
			Tag->Type |= 0x80;
		}

		WriteFile(hFile, &Tag->Type, 1, &NumWrite, NULL);

		if (Tag->sName) {
			WORD Length = strlen(Tag->sName);
			WriteFile(hFile, &Length, sizeof(WORD), &NumWrite, NULL);
			WriteFile(hFile, Tag->sName, Length, &NumWrite, NULL);
		} else {
			WriteFile(hFile, &Tag->uName, sizeof(BYTE), &NumWrite, NULL);
		}

		Tag->Type &= ~0x80;

		if (Tag->Type == TAGTYPE_BLOB) {
			WriteFile(hFile, &Tag->BlobSize, sizeof(DWORD), &NumWrite, NULL);
			WriteFile(hFile, Tag->ValPtr, Tag->BlobSize, &NumWrite, NULL);
		} else if (Tag->Type == TAGTYPE_STRING) {
			WORD Length = strlen(Tag->ValStr);
			WriteFile(hFile, &Length, sizeof(WORD), &NumWrite, NULL);
			WriteFile(hFile, Tag->ValStr, Length, &NumWrite, NULL);
		} else if (Tag->Type == TAGTYPE_HASH) {
			WriteFile(hFile, Tag->ValPtr, 16, &NumWrite, NULL);
		} else if (Tag->Type == TAGTYPE_FLOAT32) {
			WriteFile(hFile, &Tag->Valfp32, sizeof(FLOAT), &NumWrite, NULL);
		} else if (Tag->Type == TAGTYPE_UINT32) {
			WriteFile(hFile, &Tag->Valu32, sizeof(DWORD), &NumWrite, NULL);
		} else if (Tag->Type == TAGTYPE_UINT16) {
			WriteFile(hFile, &Tag->Valu16, sizeof(WORD), &NumWrite, NULL);
		} else if (Tag->Type == TAGTYPE_UINT8) {
			WriteFile(hFile, &Tag->Valu8, sizeof(BYTE), &NumWrite, NULL);
		} else if (Tag->Type == TAGTYPE_BOOL) {
			WriteFile(hFile, &Tag->Valu8, sizeof(BYTE), &NumWrite, NULL);
		} else {
			printf("Error: Unkown Tag Type %d\n", Tag->Type);
		}
		DestroyTag(Tag);
	}

	CloseHandle(hFile);
	return TRUE;
}

int main(int argc, char* argv[]) {
	char *MetFile = NULL;
	char PartFile[MAX_PATH] = {0};
	BYTE* MetFileBuffer = NULL;
	HANDLE hFile = NULL;
	DWORD NumRead, NumWrite, Idx, TagIdx;
	char StrMetSize[32] = {0};
	BOOL DoRebuild = FALSE;
	DWORD FileSize;

	DWORD MetPtr = 0, MetOldPtr = 0;
	BYTE MetVersion;
	BOOL MetNewStyle;
	EPART_FILE_FORMAT MetType;
	DWORD MetDate;
	DWORD MetTagCount;
	DWORD MetGapCount;
	DWORD MetFileSize;
	DWORD MetFreeFileSize = 0;
	CGAP *MetGaps = NULL;
	
	DWORD FsType = FSTYPE_FAT;

	if (argc < 2)  {
		printf("eMule .part file gap cleaner v1.1.\nUsage: gapcleaner <part.met file> [/rebuild]\n\n\t/rebuild clears each file gap so that eMule think download is complete.\n\teMule will than rehash the file and find corrupted parts if available.\n");
		return -1;
	}

	MetFile = argv[1];

	if (argc > 2) {
		if (StrCmpI(argv[2], "/rebuild") == 0) {
			DoRebuild = TRUE;
		}
	}

	StrCpy(PartFile, MetFile);
	PartFile[lstrlen(PartFile) - 4] = '\0';

	hFile = CreateFile(MetFile, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);

	if (hFile == INVALID_HANDLE_VALUE) {
		printf(_T("Error opening '%s' (Error %d)\n"), MetFile, GetLastError());
		goto Ende;
	}

	NumRead = GetFileSize(hFile, NULL);
	FileSize = NumRead;

	if (NumRead > 5 * 1024 * 1024) {
		printf("Error: .met file is too large, can't be loaded!\n");
		CloseHandle(hFile);
		goto Ende;
	}

	MetFileBuffer = (BYTE*) LocalAlloc(LPTR, NumRead);

	if (MetFileBuffer == NULL) {
		printf(_T("Error creating buffer for %s\n"), MetFile);
		CloseHandle(hFile);
		goto Ende;
	}
	ReadFile(hFile, MetFileBuffer, NumRead, &NumRead, NULL);
	CloseHandle(hFile);

	if (DoRebuild) {
		char OldMetFile[MAX_PATH] = {0};
		StrCpy(OldMetFile, MetFile);
		StrCat(OldMetFile, _T(".old"));
		MoveFile(MetFile, OldMetFile);
		Rebuild(MetFileBuffer, &MetPtr, MetFile);
		goto Ende;
	}
	MetVersion = *(MetFileBuffer + MetPtr++);
	
	if (MetVersion != PARTFILE_VERSION && MetVersion != PARTFILE_SPLITTEDVERSION) {
		printf("Error: .met file version %#x invalid!\n", MetVersion);
		goto Ende;
	}

	MetNewStyle = (MetVersion == PARTFILE_SPLITTEDVERSION);
	MetType = MetNewStyle ? PMT_SPLITTED : PMT_DEFAULTOLD;

	if (!MetNewStyle) {
		DWORD Internal;
		MetPtr = 24;
		Internal = *(DWORD*) (MetFileBuffer + MetPtr);
		MetPtr = 1;
		if (Internal == PARTFILE_INTERNAL) {
			MetNewStyle = TRUE;
			MetType = PMT_NEWOLD;
		}
	}

	if (MetNewStyle) {
		DWORD Temp;
		Temp = *(DWORD*) (MetFileBuffer + MetPtr);
		MetPtr += sizeof(DWORD);
		
		if (Temp == 0) {
			goto LoadHashsetFromFile;
		} else {
			BYTE Hash[16];
			MetPtr = 2;
	
			MetDate = *(DWORD*) (MetFileBuffer + MetPtr);
			MetPtr += sizeof(DWORD);

			printf("Hash Id: ");
			for (Idx = 0; Idx < 16; Idx++) {
				Hash[Idx] = *(MetFileBuffer + MetPtr + Idx);
				printf("%.2x", Hash[Idx]);
			}
			MetPtr += Idx;
			printf("\n");
		}
	} else {
		BYTE Hash[16];
		WORD MetParts;
		DWORD PartIdx;

		MetDate = *(DWORD*) (MetFileBuffer + MetPtr);
		MetPtr += sizeof(DWORD);

LoadHashsetFromFile:
		printf("Hash Id: ");
		for (Idx = 0; Idx < 16; Idx++) {
			Hash[Idx] = *(MetFileBuffer + MetPtr + Idx);
			printf("%.2x", Hash[Idx]);
		}
		printf("\n");
		MetPtr += Idx * sizeof(BYTE);
		MetParts = *(WORD*) (MetFileBuffer + MetPtr);
		MetPtr += sizeof(WORD);
		printf("Hashed Part Count: %d\n", MetParts);
		for (PartIdx = 0; PartIdx < MetParts; PartIdx++) {
//			printf("Hash Id for Part %d: ", PartIdx);
			for (Idx = 0; Idx < 16; Idx++) {
				Hash[Idx] = *(MetFileBuffer + MetPtr + Idx);
//				printf("%.2x", Hash[Idx]);
			}
//			printf("\n");
			MetPtr += Idx * sizeof(BYTE);
		}
	}
	MetTagCount = *(DWORD*) (MetFileBuffer + MetPtr);
	MetPtr += sizeof(DWORD);
	printf("Tag Count : %d\n", MetTagCount);
	
	MetGapCount = 0;
	
	MetOldPtr = MetPtr;
	for (TagIdx = 0; TagIdx < MetTagCount; TagIdx++) {
		CTAG *Tag = CreateTag(MetFileBuffer, &MetPtr);
/*		printf("Tag Type : %d\n", Tag->Type);
		printf("Tag Value : ");
		if (Tag->Type == TAGTYPE_STRING) {
			printf("'%s'\n", Tag->ValStr);
		} else if (Tag->Type == TAGTYPE_HASH) {
			for (Idx = 0; Idx < 16; Idx++) {
				printf("%.2x", ((BYTE*) Tag->ValPtr)[Idx]);
			}
			printf("\n");
		} else if (Tag->Type == TAGTYPE_UINT32) {
			printf("%.8x\n", Tag->Valu32);
		} else if (Tag->Type == TAGTYPE_FLOAT32) {
			printf("%f\n", Tag->Valfp32);
		}*/
		if (Tag->uName == 0) {
			if (Tag->sName[0] == FT_GAPSTART) {
				MetGapCount++;
			}
			if(Tag->sName[0] == FT_GAPEND) {
			}
		} else if (Tag->uName == FT_FILESIZE) {
			MetFileSize = Tag->Valu32;
		} else if (Tag->uName == FT_PARTFILENAME) {
			if (Tag->Type == TAGTYPE_STRING) {
				PathRemoveFileSpec(PartFile);
				PathAppend(PartFile, Tag->ValStr);
			}
		}
		DestroyTag(Tag);
	}
	
	StrFormatByteSize(MetFileSize, StrMetSize, sizeof(StrMetSize));
	printf("Gap Count: %d (File Size: %s)\n", MetGapCount, StrMetSize);

	MetPtr = MetOldPtr;
	MetGaps = (CGAP*) LocalAlloc(LPTR, MetGapCount * sizeof(CGAP));

	for (TagIdx = 0; TagIdx < MetTagCount; TagIdx++) {
		CTAG *Tag = CreateTag(MetFileBuffer, &MetPtr);
		if (Tag->uName == 0) {
			DWORD GapIdx;
			if (Tag->sName[0] == FT_GAPSTART) {
				GapIdx = atoi(&Tag->sName[1]);
				MetGaps[GapIdx].Start = Tag->Valu32;
			}
			if(Tag->sName[0] == FT_GAPEND) {
				GapIdx = atoi(&Tag->sName[1]);
				MetGaps[GapIdx].End = Tag->Valu32 - 1;
			}
		}
		DestroyTag(Tag);
	}

	printf("Gaps in '%s':\n", PartFile);

	hFile = CreateFile(PartFile, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);
	if (hFile == INVALID_HANDLE_VALUE) {
		printf("Can't open part file '%s' (Error %d)\n", PartFile, GetLastError());
		goto Ende;
	}
	
	FsType = DeviceIoControl(hFile, FSCTL_SET_SPARSE, NULL, 0, NULL, 0, &NumRead, NULL) ? FSTYPE_NTFS : FSTYPE_FAT;

	for (Idx = 0; Idx < MetGapCount; Idx++) {
		const DWORD BlockSize = 128 * 1024;
		BYTE DataBuffer[BlockSize];
		DWORD DataIdx = 0;
		UINT64 DataSize = 0;
		int x, y;

		char StrSize[3][32] = {0};
		DWORD GapSize = (DWORD) (MetGaps[Idx].End - MetGaps[Idx].Start);

		StrFormatByteSize((DWORD) MetGaps[Idx].Start, StrSize[0], sizeof(StrSize[0]));
		StrFormatByteSize((DWORD) MetGaps[Idx].End, StrSize[1], sizeof(StrSize[1]));
		StrFormatByteSize(GapSize, StrSize[2], sizeof(StrSize[2]));

		printf("	%s ... %s (%s)", StrSize[0], StrSize[1], StrSize[2]);
		MetFreeFileSize += GapSize;

		y = GetY();
		x = GetX();

		if (SetFilePointer(hFile, (DWORD) MetGaps[Idx].Start, NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER) {
			printf(" ... [no data]\n");
			continue;
		}

		if (FsType == FSTYPE_NTFS) {
			FILE_ZERO_DATA_INFORMATION zd = {0};
			zd.FileOffset.QuadPart = MetGaps[Idx].Start;
			zd.BeyondFinalZero.QuadPart = MetGaps[Idx].End + 1;
			DeviceIoControl(hFile, FSCTL_SET_ZERO_DATA, &zd, sizeof(zd), NULL, 0, &NumRead, NULL);
			printf(" ... Zeroing 100 %%\n");
		} else {
			while (GapSize) {
				NumRead = 0;
				ReadFile(hFile, DataBuffer, BlockSize, &NumRead, NULL);
				if (NumRead == 0) break;

				ZeroMemory(DataBuffer, Min(NumRead, GapSize));
				DataSize += Min(NumRead, GapSize);

				if (GapSize >= NumRead)
					GapSize -= NumRead;
				else
					GapSize = 0;

				SetFilePointer(hFile, -(LONG) NumRead, NULL, FILE_CURRENT);
				WriteFile(hFile, DataBuffer, NumRead, &NumWrite, NULL);
				
				SetPosition(x, y);
				printf(" ... Zeroing %3u %%", (DataSize * 100) / (MetGaps[Idx].End - MetGaps[Idx].Start));
			}
		}
		printf("\n");
	}
	
	StrFormatByteSize(MetFreeFileSize, StrMetSize, sizeof(StrMetSize));
	printf("Cleaned space in '%s': %s\n", PartFile, StrMetSize);

	CloseHandle(hFile);
Ende:
	if (MetFileBuffer) LocalFree(MetFileBuffer);
	return 0;
}

